#include <stdint.h>
#include <stdio.h>
#include <tgmath.h>
#include <timers.h>
#include <timing.h>

#include "swplatform.h"

#define PI 3.14159
#define DAC_CHAN_X 0
#define DAC_CHAN_Y 1

#undef USE_SPI_DAC

dac_t *dac;
instrument_t *ctlpanel;
ioport_t *ioport;

struct FGen {
    uint8_t lastx, lasty;
    uint8_t gen1x,gen1y,gen2x,gen2y,gen3x,gen3y;
    uint8_t mul1,mul2,mul3;
    uint32_t tim;
} fgen;

void spi_dac_pulse_clk() {
    ioport_set_value(ioport,SPIDACTEST_CLK,1);
    ioport_set_value(ioport,SPIDACTEST_CLK,0);
}

void spi_dac_pulse_load() {
    int busy = ioport_get_value(ioport,SPIDACTEST_BUSY);
    if (busy>0) {
        printf("Warning: loading although BUSY\n");
    }
    ioport_set_value(ioport,SPIDACTEST_LOAD,1);
    spi_dac_pulse_clk();
    ioport_set_value(ioport,SPIDACTEST_LOAD,0);
}

void spi_dac_put(uint16_t val) {
    printf("DAC_PUT: setting data value %04x\n",val);
    ioport_set_value(ioport,SPIDACTEST_DATAOUT,val);
    spi_dac_pulse_load();
    int busy = ioport_get_value(ioport,SPIDACTEST_BUSY);
    if (!busy) printf("Error: not busy\n");
    int cnt = 128;
    while (cnt>0) {
        spi_dac_pulse_clk();
        cnt--;
        busy = ioport_get_value(ioport,SPIDACTEST_BUSY);
        if (!busy) break;
    }
    if (busy) {
        printf("Error: still busy after 128 clock pulses!\n");
    } else {
        printf("BUSY returned to OFF after %d pulses\n",128-cnt);
    }
}

void set_dac_chan(uint8_t chan, uint8_t val, bool flush) {
    spi_dac_put((chan<<14)+(flush?(1<<12):0)+(val<<4)+0);
}

void dacset(uint8_t chan, uint8_t val, bool flush) {
#ifdef USE_SPI
    dac_write(dac,chan,val,flush);
#else
    set_dac_chan(chan,val,flush);
#endif
}

void genloop(struct FGen fg) {
    uint8_t xv, yv;
    bool upd;
    fg.tim++;
    fg.tim = fg.tim % 1000;
    double phi;
    /* Generator 1 - sawtooth */
    fg.gen1x = (fg.gen1x+1) % 256;
    fg.gen1y = (fg.gen1y+3) % 256;
    /* Generator 2 - Sine / Cosine */
    phi = (fg.tim%1000*2*PI)/1000;
    fg.gen2x = 128*(1.0+(sin(phi*2)));
    fg.gen2y = 128*(1.0+(cos(phi)));
    /* Generator 3 is manual (no code here) */
    /* now comes the mixer */
    xv = (fg.mul1*fg.gen1x) + (fg.mul2*fg.gen2x) + (fg.mul3*fg.gen3x);
    yv = (fg.mul1*fg.gen1y) + (fg.mul2*fg.gen2y) + (fg.mul3*fg.gen3y);
    upd = false;
    if ((xv!=fg.lastx) || (yv!=fg.lasty)) {
         upd = true;
         fg.lastx = xv;
         fg.lasty = yv;
    }
    if (upd) {
       /* printf("X=%d Y=%d\n",xv,yv); */
       instrument_set_value(ctlpanel,CTLPANEL_XOUT,xv);
       instrument_set_value(ctlpanel,CTLPANEL_YOUT,yv);
       dacset(DAC_CHAN_X,xv,false);
       dacset(DAC_CHAN_Y,yv,true);
    }
}

void init() {
    dac = dac_open(DAC_1);
    ctlpanel = instrument_open(CTLPANEL);
    ioport = ioport_open(SPIDACTEST);
    fgen.mul1 = 0;
    fgen.mul2 = 0;
    fgen.mul3 = 1;
    fgen.tim = 0;
    timer_register_handler(&fgen,1000,(timer_handler_t) genloop);
}

void main() {
    uint8_t sel_saw,sel_sine,sel_man;
    init();
    uint64_t fhz = freq_hz();
    printf("NOS031004 started, clock frequency is %d Hz\n",freq_hz());
    for (;;) {
        fgen.gen3x = instrument_get_value(ctlpanel,CTLPANEL_XIN);
        fgen.gen3y = instrument_get_value(ctlpanel,CTLPANEL_YIN);
        sel_saw = instrument_get_value(ctlpanel,CTLPANEL_SEL_SAW);
        sel_sine = instrument_get_value(ctlpanel,CTLPANEL_SEL_SINE);
        sel_man = instrument_get_value(ctlpanel,CTLPANEL_SEL_MAN);
        fgen.mul1 = (sel_saw!=0)?1:0;
        fgen.mul2 = (sel_sine!=0)?1:0;
        fgen.mul3 = (sel_man!=0)?1:0;
    }
}

